#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#       moon.py
#
#       Copyright 2009 Espen Rønnevik <brightside@quasinet.org>
#
#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation; either version 2 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, write to the Free Software
#       Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#       MA 02110-1301, USA.

from __future__ import division
from math import sqrt, sin, cos, tan, atan, radians, degrees, fabs

# Astronomical constants.

unixepoch = 2440587.5		# Midnight 1 jan 1970
astroepoch = 2444238.5 		# Midnight 1 jan 1980
epsilon = 1.0e-6			# Limiting value for Kepler equation

# Constants defining the Sun's apparent orbit.

s_elong = 278.833540		# ecliptic longitude of the Sun at epoch
s_elongp = 282.596403		# ecliptic longitude of the Sun at perigee
s_eccent = 0.016718			# eccentricity of Earth's orbit

# Constants defining the Moon's orbit.

m_mlong = 64.975464			# moon's mean longitude at the epoch
m_mlongp = 349.383063		# mean longitude of the perigee at the epoch
m_longnode = 151.950429		# mean longitude of the node at the epoch
m_eccent = 0.054900			# eccentricity of the Moon's orbit
m_synodic = 29.53058868		# synodic month (new Moon to new Moon)

class Phase(object):

	def __init__(self, unixts):

		# Calculation of the Sun's position.
		# ----------------------------------

		# Modified Julian Date within astroepoch
		Day = (unixepoch + (unixts / 86400)) - astroepoch

		# Mean anomaly of the Sun
		N = self._fixangle((360.0 / 365.2422) * Day)

		# Convert from perigee coordinates to epoch
		M = self._fixangle(N + s_elong - s_elongp)

		# Solve equation of Kepler
		Ec = self._kepler(M, s_eccent)

		# True anomaly in degrees
		Ec = sqrt((1 + s_eccent) / (1 - s_eccent)) * tan(Ec / 2)
		Ec = 2 * degrees(atan(Ec))

		# Sun's geocentric ecliptic longitude
		lsun = self._fixangle(Ec + s_elongp)

		# Calculation of the Moon's position
		# ----------------------------------

		# Moon's mean longitude
		ml = self._fixangle(13.1763966 * Day + m_mlong)

		# Moon's mean anomaly
		MM = self._fixangle(ml - 0.1114041 * Day - m_mlongp)

		# Ascending node mean anomaly
		MN = self._fixangle(m_longnode - 0.0529539 * Day)

		# Evection
		Ev = 1.2739 * sin(radians(2 * (ml - lsun) - MM))

		# Annual equation.
		Ae = 0.1858 * sin(radians(M))

		# Correction term.
		A3 = 0.37 * sin(radians(M))

		# Corrected anomaly.
		MmP = MM + Ev - Ae - A3

		# Correction for the equation of the centre.
		mEc = 6.2886 * sin(radians(MmP))

		# Another correction term.
		A4 = 0.214 * sin(radians(2 * MmP))

		# Corrected longitude.
		lP = ml + Ev + mEc - Ae + A4

		# Variation.
		V = 0.6583 * sin(radians(2 * (lP - lsun)))

		# True longitude.
		lPP = lP + V

		# Calculation of the phase of the Moon.
		# -------------------------------------

		# Age of the Moon in degrees.
		agedegrees = lPP - lsun

		# Fraction of moon phase
		phase = radians(self._fixangle(agedegrees))

		# Fraction of moon illuminated.
		illuminated = (1 - cos(radians(agedegrees))) / 2

		# Moon age in days
		age = m_synodic * (self._fixangle(agedegrees) / 360.0)

		# Calculate Unix timestamp for next moonphase change
		changefraction = 1 - (age % 1)
		phasehours = int(24 * changefraction)
		phasemins = int(1440 * changefraction) % 60
		phasesecs = int(86400 * changefraction) % 60

		changesecs = phasehours * 60 * 60 + phasemins * 60 + phasesecs
		next = unixts + changesecs

		# Populate instance with calculated phase values
		self.phase = phase
		self.illuminated = illuminated
		self.age = age
		self.next = next

	def _fixangle(self, angle):
		return angle - 360 * (angle // 360);

	def _kepler(self, m, eccent):
		e = m = radians(m)
		delta = 1

		while fabs(delta) > epsilon:
			delta = e - eccent * sin(e) - m
			e = e - (delta / (1 - eccent * cos(e)))
		return e

#-----------------------------------------------------------------------

if __name__ == '__main__':
	pass
