#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#       wcz.py
#
#       Copyright 2009 Espen Rønnevik <brightside@quasinet.org>
#
#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation; either version 2 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, write to the Free Software
#       Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#       MA 02110-1301, USA.

r""" wcz - Manipulaton of WallpaperClock files.

This module allows you to load *.wcz files and generate wallpapers
from them.

Usage overview:
  Make an instance of the class, passing in the path for the wcz file
  you'd like to use and use a combination of the following two methods:

  properties:
    fout - Specify the filename of the written file.

  generate(unixtimestamp):
    This function will generate a new wallpaper from the given timestamp
    and store it in a buffer.

  Then call generate() with a unix timestamp to make
  a new wallpaper image using the timestamp information and buffer it.
  Finally call write() with a path which will write the buffered wallpaper
  to a file. Optionally call write() directly
"""

from __future__ import division

import sys
import os

from time import time, localtime, struct_time, mktime
from zipfile import ZipFile
from ConfigParser import RawConfigParser

from PythonMagick import Blob as imBlob
from PythonMagick import Image as imImage
from PythonMagick._PythonMagick import CompositeOperator as imOp

from moon import Phase

class WCZConfig(object):
    """
    Special "file-like" object to handle configuration files.

    ConfigParser is sensitive to newline-endings and the zipfile module does
    not handle them correctly, even if pythons universal newline-endings
    feature is specified.

    This class reads in the configuration file clock.ini from a WCZfile,
    fixes newline-endings and allows them to be retrieved by readline()
    """
    def __init__(self, zfile):
        buf = zfile.open('clock.ini', 'r').read()
        buf = buf.replace('\r\n', '\n').replace('\r', '\n')
        self.buffer = buf.split('\n')

    def readline(self, size = -1):
        try:
            return self.buffer.pop(0) + '\n'
        except IndexError:
            return ''

class WCZ(object):

    def __init__(self, wczname, use12h):
        """
        Create a new wallpaperclock instance.
        Accepted input is the full path to the wcz-file and full path
        to a file where generated wallpapers should be written.
        """

        f = ZipFile(wczname, 'r')

        # Read all the images, create mapping from names and blobs
        wpnames = [n for n in f.namelist() if not n == 'clock.ini' or
            n[:7] == 'preview']
        wpblobs = [imBlob(f.open(n).read()) for n in wpnames]
        self.wpimages = dict(zip(wpnames, wpblobs))

        # Read the clock configuration file
        config = RawConfigParser()
        config.readfp(WCZConfig(f))

        f.close()

        cachenames = ('nextdgen', 'day.png', 'hour.png')
        cachevals = (0, imBlob(), imBlob())
        self.cache = dict(zip(cachenames, cachevals))

        self.wpbuffer = None
        self.settings = {}
        self.settings['name'] = config.get('Settings', 'name')
        self.settings['height'] = config.getint('Settings', 'height')
        self.settings['width'] = config.getint('Settings', 'width')
        self.settings['hrefresh'] = config.getint(
            'Settings', 'refreshhourinterval')
        self.settings['himages'] = config.getint('Settings', 'hourimages')

        # Everything except the background image, preview images and
        # configuration file is optional, so check what is available
        self.settings['haveZodiac'] = bool([i for i in wpnames if
            i[:6] == 'zodiac'])
        self.settings['haveMoonphase'] = bool([i for i in wpnames if
            i[:9] == 'moonphase'])
        self.settings['haveMonth'] = bool([i for i in wpnames if
            i[:5] == 'month'])
        self.settings['haveWeekday'] = bool([i for i in wpnames if
            i[:7] == 'weekday'])
        self.settings['haveDay'] = bool([i for i in wpnames if
            i[:3] == 'day'])
        self.settings['haveHour'] = bool([i for i in wpnames if
            i[:4] == 'hour'])
        self.settings['haveMinute'] = bool([i for i in wpnames if
            i[:6] == 'minute'])
        self.settings['have12hSymbol'] = config.getboolean(
            'Settings', 'ampmenabled') and use12h

        # Switch to 12hour clockmode if the 12hSymbol is active or if the
        # number of images indicates this is a 12hour clock
        self.settings['12hMode'] = self.settings['have12hSymbol'] or (
            (self.settings['himages'] * self.settings['hrefresh']) / 60 == 12)

    def getZodiac(self, ts, tts):
        """ Returns the correct zodiac imagename """

        if tts.tm_mon == 1:
            if tts.tm_mday < 20:
                zodiac = 'zodiacCapricorn.png'
            else:
                zodiac = 'zodiacAquarius.png'
        elif tts.tm_mon == 2:
            if tts.tm_mday < 18:
                zodiac = 'zodiacCapricorn.png'
            else:
                zodiac = 'zodiacPisces.png'
        elif tts.tm_mon == 3:
            if tts.tm_mday < 20:
                zodiac = 'zodiacPisces.png'
            else:
                zodiac = 'zodiacAries.png'
        elif tts.tm_mon == 4:
            if tts.tm_mday < 19:
                zodiac = 'zodiacAries.png'
            else:
                zodiac = 'zodiacTaurus.png'
        elif tts.tm_mon == 5:
            if tts.tm_mday < 20:
                zodiac = 'zodiacTaurus.png'
            else:
                zodiac = 'zodiacGemini.png'
        elif tts.tm_mon == 6:
            if tts.tm_mday < 21:
                zodiac = 'zodiacGemini.png'
            else:
                zodiac = 'zodiacCancer.png'
        elif tts.tm_mon == 7:
            if tts.tm_mday < 22:
                zodiac = 'zodiacCancer.png'
            else:
                zodiac = 'zodiacLeo.png'
        elif tts.tm_mon == 8:
            if tts.tm_mday < 23:
                zodiac = 'zodiacLeo.png'
            else:
                zodiac = 'zodiacVirgo.png'
        elif tts.tm_mon == 9:
            if tts.tm_mday < 22:
                zodiac = 'zodiacVirgo.png'
            else:
                zodiac = 'zodiacLibra.png'
        elif tts.tm_mon == 10:
            if tts.tm_mday < 23:
                zodiac = 'zodiacLibra.png'
            else:
                zodiac = 'zodiacScorpio.png'
        elif tts.tm_mon == 11:
            if tts.tm_mday < 22:
                zodiac = 'zodiacScorpio.png'
            else:
                zodiac = 'zodiacSagittarius.png'
        else: # tts.tm_mon == 12
            if tts.tm_mday < 21:
                zodiac = 'zodiacSagittarius.png'
            else:
                zodiac = 'zodiacCapricorn.png'

        return zodiac

    def getMoonphase(self, ts, tts):
        """
        Returns a tuple with the correct moonphase imagename
        and unix timestamp of next moonphase change
        """

        mp = Phase(ts)

        # Set the next cache generation time to next moon phase
        self.cache['nextdgen'] = mp.next

        # Shift age up by one, moonphase-images go from 1 to 30
        phaseimage = 'moonphase%s.png' % (int(mp.age + 1),)

        return phaseimage

    def getMonth(self, ts, tts):
        """ Returns the correct month imagename """

        month = 'month%d.png' % (tts.tm_mon,)
        return month

    def getWeekday(self, ts, tts):
        """ Returns the correct weekday imagename """

        weekday = 'weekday%d.png' % (tts.tm_wday + 1,)
        return weekday

    def getDay(self, ts, tts):
        """ Returns the correct day imagename """

        day = 'day%d.png' % (tts.tm_mday,)
        return day

    def getHour(self, ts, tts):
        """ Returns the correct hour imagename """

        # Calculating the hour is more complicated, because we need to
        # factor in the hour refreshrate (once every 12 or 60 minutes)
        # and also if we're using a 12hour clock or a 24hour one.

        # While not strictly required yet, this function should be able
        # to handle any hourly refreshrate as long as there are
        # clockhours * (60/refresh) images available.

        if self.settings['12hMode'] and tts.tm_hour > 11:
            hcount = (tts.tm_hour-12) * (60 // self.settings['hrefresh'])
        else:
            hcount = tts.tm_hour * (60 / self.settings['hrefresh'])
        mcount = int(tts.tm_min / self.settings['hrefresh'])

        hour = 'hour%d.png' % (hcount + mcount,)
        return hour

    def getMinute(self, ts, tts):
        """ Return the correct minute imagename """

        minute = 'minute%d.png' % (tts.tm_min,)
        return minute

    def get12hSymbol(self, ts, tts):
        """ Return the correct 12hsymbol imagename """

        if tts.tm_hour > 12:
            symbol = 'pm.png'
        else:
            symbol = 'am.png'

        return symbol

    def generate_dcache(self, ts, tts, op):
        overlays = [self.__getattribute__('get'+n)(ts, tts) for n in
          ('Zodiac', 'Moonphase', 'Month', 'Weekday', 'Day') if
          self.settings['have'+n]]
        dcache = imImage(self.wpimages[overlays[0]])

        for overlay in overlays[1:]:
            if overlay in self.wpimages.keys():
                dcache.composite(imImage(self.wpimages[overlay]), 0, 0, op)

        dcache.write(self.cache['day.png'])

        # Compute unix timestamp for midnight today
        midnight = mktime(struct_time(
          (tts[0], tts[1], tts[2], 0, 0, 0, tts[6], tts[7], tts[8])))

        # Shift up to get timestamp for midnight tomorrow
        nextmidnight = midnight + 84600

        # Set next daycache generation time to whichever comes first,
        # next midnight or next moonphase change
        self.cache['nextdgen'] = min(nextmidnight, self.cache.get(
            next, nextmidnight))

    def generate_hcache(self, ts, tts, op):
        if self.settings['haveHour']:
            hoverlay = imImage(self.wpimages[self.getHour(ts, tts)])
            hcache = imImage(self.cache['day.png'])
            hcache.composite(hoverlay, 0, 0, op)
            hcache.write(self.cache['hour.png'])
        else:
            self.cache['hour.png'] = self.cache['day.png']

    def generate(self, ts):
        """
        Generates a new wallpaper image and store it in a buffer.
        Input is Unix timestamp for the date & time of the generated
        wallpaper image.
        """

        tts = localtime(ts)
        op = imOp().OverCompositeOp

        # Check to see if we should generate a new cached image. We use two
        # cached images, dcache and hcache. dcache is constructed out of
        # Zodiac, Moonphase, Month, Weekday and Day images and must be
        # regenerated every midnight and noon. hcache is dcache combined with
        # an hour image and is regenerated every hour.

        if ts >= self.cache['nextdgen']:
            self.generate_dcache(ts, tts, op)
            self.generate_hcache(ts, tts, op)
        elif tts.tm_min == 0:
            self.generate_hcache(ts, tts, op)

        # Grab overlay images needed to build final image
        overlays = [self.__getattribute__('get'+n)(ts, tts) for n in
          ('Minute', '12hSymbol') if self.settings['have'+n]]

        # Build the image, starting with the background and hcache overlay
        wp = imImage(self.wpimages['bg.jpg'])
        wp.composite(imImage(self.cache['hour.png']), 0, 0, op)

        # Then comes the overlays in the specified order.
        for overlay in overlays:
            if overlay in self.wpimages.keys():
                wp.composite(imImage(self.wpimages[overlay]), 0, 0, op)

        self.wpbuffer = wp

    def write(self, f = None):
        """
        Write the wallpaper buffer to the path given, or if no path is
        specified, use the path from self.fout instead.

        If no wallpaper have been generated and the buffer is empty,
        it will generate a new wallpaper using current time and write
        that to file.
        """

        if f is None:
            f = self.fout

        if self.wpbuffer is None:
            self.generate(time())

        self.wpbuffer.write(f)
        self.wpbuffer = None

#-----------------------------------------------------------------------

if __name__ == '__main__':
    pass
